from __future__ import print_function

import argparse
import json
import os
import telebot
import warnings

from flask import Flask, request

from .dialog_connector import DialogConnector, SOURCES
from .message_logging import LoggedMessage


class FlaskServer:
    def __init__(
            self,
            connector: DialogConnector,
            telegram_token=None, base_url=None,
            alice_url='alice/', telegram_url='tg/', restart_webhook_url='restart_webhook',
            collection_for_logs=None
    ):
        if telegram_token is None:
            telegram_token = os.environ.get('TOKEN')
        self.telegram_token = telegram_token
        if base_url is None:
            base_url = os.environ.get('BASE_URL')
        self.base_url = base_url
        self.alice_url = alice_url
        self.telegram_url = telegram_url
        self.restart_webhook_url = restart_webhook_url

        self.connector = connector
        self.collection_for_logs = collection_for_logs

        self.app = Flask(__name__)

        self.app.route("/" + self.alice_url, methods=['POST'])(self.alice_response)
        if self.telegram_token is not None:
            self.bot = telebot.TeleBot(telegram_token)
            self.bot.message_handler(func=lambda message: True)(self.tg_response)
            self.app.route('/' + self.telegram_url + self.telegram_token, methods=['POST'])(self.get_tg_message)
            self.app.route("/" + self.restart_webhook_url)(self.telegram_web_hook)
        else:
            self.bot = None

    def alice_response(self):
        if self.collection_for_logs is not None:
            LoggedMessage.from_alice(request.json).save_to_mongo(self.collection_for_logs)
        response = self.connector.respond(request.json, source=SOURCES.ALICE)
        if self.collection_for_logs is not None:
            LoggedMessage.from_alice(response).save_to_mongo(self.collection_for_logs)
        return json.dumps(response, ensure_ascii=False, indent=2)

    def tg_response(self, message):
        if self.collection_for_logs is not None:
            LoggedMessage.from_telegram(message).save_to_mongo(self.collection_for_logs)
        response = self.connector.respond(message, source=SOURCES.TELEGRAM)
        telegram_response = self.bot.reply_to(message, **response)
        if self.collection_for_logs is not None:
            LoggedMessage.from_telegram(telegram_response).save_to_mongo(self.collection_for_logs)

    def get_tg_message(self):
        self.bot.process_new_updates([telebot.types.Update.de_json(request.stream.read().decode("utf-8"))])
        return "!", 200

    def telegram_web_hook(self):
        self.bot.remove_webhook()
        self.bot.set_webhook(url=self.base_url + self.telegram_url + self.telegram_token)
        return "Weebhook restarted!", 200

    def run_local_telegram(self):
        if self.bot is not None:
            self.bot.polling()
        else:
            raise ValueError('Cannot run Telegram bot, because Telegram token was not found.')

    def run_server(self, host="0.0.0.0", port=None):
        if self.telegram_token is not None:
            self.telegram_web_hook()
        else:
            warnings.warn('Telegram token was not found, running for Alice only.')
        if port is None:
            port = int(os.environ.get('PORT', 5000))
        self.app.run(host=host, port=port)

    def run_command_line(self):
        input_sentence = ''
        while True:
            response, need_to_exit = self.connector.respond(input_sentence, source=SOURCES.TEXT)
            print(response)
            if need_to_exit:
                break
            input_sentence = input('> ')

    def parse_args_and_run(self):
        parser = argparse.ArgumentParser(description='Run the bot')
        parser.add_argument('--cli', action='store_true', help='Run the bot locally in command line mode')
        parser.add_argument('--poll', action='store_true', help='Run the bot locally in polling mode (Telegram only)')
        args = parser.parse_args()
        if args.cli:
            self.run_command_line()
        elif args.poll:
            self.run_local_telegram()
        else:
            self.run_server()



# coding: utf-8

from ..nlu import basic_nlu


class COMMANDS:
    EXIT = 'exit'


class Context:
    def __init__(self, user_object, message_text, metadata):
        self.user_object = user_object
        self.message_text = message_text
        self.metadata = metadata
        # todo: compare old and new user objects


class Response:
    def __init__(self, response, suggests=None, commands=None):
        pass
    # todo: use this class instead of 4 outputs


class BaseDialogManager:
    """ This class defines the interface of dialog managers with a single `response` function. """
    def __init__(self, fixed_response='I always answer with this text.', fixed_suggests=None):
        self.fixed_response = fixed_response
        self.fixed_suggests = fixed_suggests or []

    def respond(self, user_object, message_text, metadata):
        updated_user_object = user_object
        response = self.fixed_response
        suggests = self.fixed_suggests
        commands = []
        return updated_user_object, response, suggests, commands


class CascadableDialogManager(BaseDialogManager):
    """ This interface allows dialog manager to have no answer - and to let other managers try.
    The `method try_to_respond` is like `respond`, but may also return None.
    It is expected to be used with CascadeDialogManager.
    """
    def __init__(self, *args, default_message='I don\'t understand', **kwargs):
        super(CascadableDialogManager, self).__init__(*args, **kwargs)
        self.default_message = default_message

    def try_to_respond(self, *args, **kwargs):
        raise NotImplementedError()

    def respond(self, user_object, message_text, metadata):
        response = self.try_to_respond(user_object, message_text, metadata)
        if response is not None:
            return response
        return user_object, self.default_message, [], []


class CascadeDialogManager(BaseDialogManager):
    """ This dialog manager tries multiple dialog managers in turn, and returns the first successful response. """
    def __init__(self, *managers):
        super(CascadeDialogManager, self).__init__()
        assert len(managers) > 0
        for manager in managers[:-1]:
            assert isinstance(manager, CascadableDialogManager)
        self.candidates = managers[:-1]
        assert isinstance(managers[-1], BaseDialogManager)
        self.final_candidate = managers[-1]

    def respond(self, *args, **kwargs):
        for manager in self.candidates:
            response = manager.try_to_respond(*args, **kwargs)
            if response is not None:
                return response
        return self.final_candidate.respond(*args, **kwargs)


class GreetAndHelpDialogManager(CascadableDialogManager):
    """ This dialog manager can be responsible for the first and the last messages, and for the help message. """
    def __init__(self, greeting_message, help_message, *args, exit_message=None, **kwargs):
        super(GreetAndHelpDialogManager, self).__init__(*args, **kwargs)
        self.greeting_message = greeting_message
        self.help_message = help_message
        self.exit_message = exit_message

    def try_to_respond(self, user_object, message_text, metadata):
        context = Context(user_object=user_object, message_text=message_text, metadata=metadata)
        if self.is_first_message(context):
            return user_object, self.greeting_message, [], []
        if self.is_like_help(context):
            return user_object, self.help_message, [], []
        if self.exit_message is not None and self.is_like_exit(context):
            return user_object, self.exit_message, [], [COMMANDS.EXIT]
        return None

    def is_first_message(self, context):
        if not context.message_text or context.message_text == '/start':
            return True
        if basic_nlu.like_help(context.message_text):
            return True
        return False

    def is_like_help(self, context):
        return basic_nlu.like_help(context.message_text)

    def is_like_exit(self, context):
        return basic_nlu.like_exit(context.message_text)



import re


def fast_normalize(text):
    text = re.sub('[^a-zа-я0-9\-]+', ' ', text.lower())
    text = re.sub('\s+', ' ', text).strip()
    return text


def like_help(text):
    text = fast_normalize(text)
    return bool(re.match('^(алиса |яндекс )?(помощь|что ты (умеешь|можешь))$', text))


def like_exit(text):
    text = fast_normalize(text)
    return bool(re.match('^(алиса |яндекс )?(выход|хватит( болтать| играть)?|выйти|закончить)$', text))


def like_yes(text):
    text = fast_normalize(text)
    return bool(re.match('^(да|ага|окей|ок|конечно|yes|yep|хорошо|ладно)$', text))


def like_no(text):
    text = fast_normalize(text)
    return bool(re.match('^(нет|не|no|nope)$', text))



from telebot import types as teletypes
from datetime import datetime


class LoggedMessage:
    def __init__(self, text, user_id, from_user, **kwargs):
        self.text = text
        self.user_id = user_id
        self.from_user = from_user
        self.timestamp = str(datetime.utcnow())
        self.kwargs = kwargs

    def save_to_mongo(self, collection):
        collection.insert_one(self.to_dict())

    def to_dict(self):
        result = {
            'text': self.text,
            'user_id': self.user_id,
            'from_user': self.from_user,
            'timestamp': self.timestamp
        }
        for k, v in self.kwargs.items():
            if k not in result:
                result[k] = v
        return result

    @classmethod
    def from_telegram(cls, message: teletypes.Message, **kwargs):
        reply_to_id = None
        if message.reply_to_message is not None:
            reply_to_id = message.reply_to_message.message_id
        return cls(
            text=message.text,
            user_id=message.chat.id,
            message_id=message.message_id,
            from_user=not message.from_user.is_bot,
            username=message.chat.username,
            reply_to_id=reply_to_id,
            source='telegram',
            **kwargs
        )

    @classmethod
    def from_alice(cls, message, **kwargs):
        return cls(
            text=message['response']['text'] if 'response' in message else message['request']['original_utterance'],
            user_id=message['session']['user_id'],
            from_user='response' not in message,
            data=message,
            source='alice',
            **kwargs
        )



import math
import textdistance

from collections import Counter

from ..nlu import basic_nlu


class BaseMatcher:
    def __init__(self, threshold=0.5):
        self.threshold = threshold

    def fit(self, texts, labels):
        raise NotImplementedError()

    def match(self, text):
        scores, labels = self.get_scores(text)
        best_score = 0.0
        winner_label = None
        for score, label in zip(scores, labels):
            if score >= self.threshold and score > best_score:
                best_score = score
                winner_label = label
        return winner_label, best_score

    def get_scores(self, text):
        raise NotImplementedError()


class PairwiseMatcher(BaseMatcher):
    def __init__(self, *args, text_normalization='fast', **kwargs):
        super(PairwiseMatcher, self).__init__(*args, **kwargs)
        self.text_normalization = text_normalization
        self._texts = []
        self._labels = []

    def preprocess(self, text):
        if self.text_normalization == 'fast':
            text = basic_nlu.fast_normalize(text)
        return text

    def compare(self, one, another):
        raise NotImplementedError()

    def fit(self, texts, labels):
        self._texts = [self.preprocess(text) for text in texts]
        self._labels = labels

    def get_scores(self, text):
        processed = self.preprocess(text)
        return [self.compare(processed, t) for t in self._texts], self._labels


class ExactMatcher(PairwiseMatcher):
    def compare(self, one, another):
        return float(one == another)


class TextDistanceMatcher(PairwiseMatcher):
    def __init__(self, *args, by_words=True, metric='cosine', **kwargs):
        super(TextDistanceMatcher, self).__init__(*args, **kwargs)
        self.by_words = by_words
        self.metric = metric
        self.fun = getattr(textdistance, metric).normalized_similarity

    def preprocess(self, text):
        text = super(TextDistanceMatcher, self).preprocess(text)
        if self.by_words:
            return text.split()
        return text

    def compare(self, one, another):
        return self.fun(one, another)


class JaccardMatcher(PairwiseMatcher):
    def preprocess(self, text):
        text = super(JaccardMatcher, self).preprocess(text)
        return set(text.split())

    def compare(self, one, another):
        intersection = len(one.intersection(another))
        union = len(one.intersection(another))
        if intersection:
            return intersection / union
        return 0.0


class TFIDFMatcher(PairwiseMatcher):
    def __init__(self, *args, smooth=2.0, ngram=1, **kwargs):
        super(TFIDFMatcher, self).__init__(*args, **kwargs)
        self.smooth = smooth
        self.ngram = ngram
        self.vocab = Counter()

    def fit(self, texts, labels):
        self.vocab = Counter(w for t in texts for w in self._tokenize(t))
        self._texts = [self.preprocess(text) for text in texts]
        self._labels = labels

    def preprocess(self, text):
        text = super(TFIDFMatcher, self).preprocess(text)
        tf = Counter(self._tokenize(text))
        return {w: tf / math.log(self.smooth + self.vocab[w]) for w, tf in tf.items()}

    def compare(self, one, another):
        dot = self._dot(one, another)
        if dot < 1e-6:
            return 0.0
        return dot / math.sqrt(self._norm(one) * self._norm(another))

    def _tokenize(self, text):
        words = text.split()
        if self.ngram == 1:
            return words
        words = ['BOS'] + words + ['EOS']
        return words + ['_'.join(words[i:(i + self.ngram)]) for i in range(len(words) - self.ngram + 1)]

    def _dot(self, one, another):
        return sum(v * another.get(k, 0) for k, v in one.items())

    def _norm(self, one):
        return self._dot(one, one)


_matchers = dict()


def register_matcher(name, matcher_maker):
    _matchers[name] = matcher_maker


def make_matcher(key, **kwargs):
    return _matchers[key](**kwargs)


register_matcher('exact', lambda **kwargs: ExactMatcher(**kwargs))
register_matcher('levenshtein', lambda **kwargs: TextDistanceMatcher(by_words=False, metric='levenshtein', **kwargs))
register_matcher('cosine', lambda **kwargs: TextDistanceMatcher(by_words=True, metric='cosine', **kwargs))
register_matcher('tf-id', lambda **kwargs: TFIDFMatcher(**kwargs))



import re
import yaml

from collections import Mapping

from .base import CascadableDialogManager, Context
from ..nlu import basic_nlu


class FieldConfig:
    def __init__(self, obj, idx=-1):
        self.validate_regexp = re.compile(obj.get('validate_regexp', '.*'))
        self.id = id
        self.name = obj.get('name', 'field_{}'.format(idx))
        self.question = obj['question']
        self.validate_message = obj.get('validate_message')
        self.options = obj.get('options')
        self.suggests = obj.get('suggests')


class FormConfig:
    def __init__(self, config):
        if isinstance(config, str):
            with open(config, 'r', encoding='utf-8') as f:
                self._cfg = yaml.load(f)
        elif isinstance(config, Mapping):
            self._cfg = config
        else:
            raise ValueError('Config should be a yaml filename or a dict')
        # todo: validate everything
        self.form_name = self._cfg['form_name']

        self.start_regex = re.compile(self._cfg['start'].get('regexp', '.*'))
        self.start_message = self._cfg['start'].get('message')
        self.start_suggests = self._cfg['start'].get('suggests', [])

        exit_block = self._cfg.get('exit', {})
        self.exit_regexp = re.compile(exit_block['regexp']) if 'regexp' in exit_block else None
        self.exit_message = exit_block.get('message')
        self.exit_suggest = exit_block.get('suggest')

        self.finish_message = self._cfg.get('finish', {}).get('message')

        self.fields = [FieldConfig(obj, idx=i) for i, obj in enumerate(self._cfg['fields'])]
        self.default_field = FieldConfig(self._cfg['default_field']) if 'default_field' in self._cfg else None
        self.num_fields = len(self.fields)


class FormFillingDialogManager(CascadableDialogManager):
    def __init__(self, config, *args, **kwargs):
        super(FormFillingDialogManager, self).__init__(*args, **kwargs)
        self.config = FormConfig(config)

    def try_to_respond(self, user_object, message_text, metadata):
        context = Context(user_object=user_object, message_text=message_text, metadata=metadata)
        normalized = basic_nlu.fast_normalize(message_text)
        form = user_object.get('forms', {}).get(self.config.form_name, {})
        if form.get('is_active'):
            if self.config.exit_regexp and re.match(self.config.exit_regexp, normalized):
                form['is_active'] = False
                return user_object, self.config.exit_message, [], []
            question_id = form['next_question']
            if self.answer_is_valid(form, message_text):
                form['fields'][self.config.fields[question_id].name] = message_text
                next_question_id = question_id + 1
                if next_question_id >= self.config.num_fields:
                    form['is_active'] = False
                    result = self.handle_completed_form(form, context)
                    if result is not None:
                        return result
                    return user_object, self.config.finish_message, [], []
                form['next_question'] = next_question_id
                return self.ask_question(next_question_id, user_object=user_object, reask=False)
            else:
                return self.ask_question(question_id, user_object=user_object, reask=True)
        elif re.match(self.config.start_regex, normalized):
            if 'forms' not in user_object:
                user_object['forms'] = {}
            # todo: if there is no start message, move directly to question 0
            user_object['forms'][self.config.form_name] = {
                'fields': {},
                'is_active': True,
                'next_question': -1
            }
            form = user_object['forms'][self.config.form_name]
            if self.config.start_message is not None:
                form['next_question'] = -1
                return user_object, self.config.start_message, self.config.start_suggests, []
            else:
                form['next_question'] = 0
                return self.ask_question(0, user_object, reask=False)
        return None

    def ask_question(self, next_question_id, user_object, reask=False):
        the_question = self.config.fields[next_question_id]
        response = the_question.question
        if reask:
            if the_question.validate_message is not None:
                response = the_question.validate_message
            elif 'validate_message' in self.config.default_field:
                response = self.config.default_field['validate_message']
        if the_question.options is not None:
            suggests = the_question.options
        elif the_question.suggests is not None:
            suggests = the_question.suggests
        else:
            suggests = []
        if self.config.exit_suggest is not None:
            suggests.append(self.config.exit_suggest)
        return user_object, response, suggests, []

    def answer_is_valid(self, form, message_text):
        question_id = form.get('next_question', -1)
        if question_id == -1:
            return True
        the_question = self.config.fields[question_id]
        if the_question.options is not None:
            return message_text in the_question.options
        if the_question.validate_regexp is not None:
            return re.match(the_question.validate_regexp, basic_nlu.fast_normalize(message_text))
        return True

    def handle_completed_form(self, form, context):
        pass



import copy
import telebot

from .session_storage import BaseStorage


class SOURCES:
    TELEGRAM = 'telegram'
    ALICE = 'alice'
    TEXT = 'text'
    unknown_source_error_message = 'Source must be on of {"telegram", "alice", "text"}'


class DialogConnector:
    COMMAND_EXIT = 'exit'
    """ This class provides unified interface for both Telegram and Alice applications """
    def __init__(self, dialog_manager, storage=None, default_source='telegram', tg_suggests_cols=1):
        self.dialog_manager = dialog_manager
        self.default_source = default_source
        self.storage = storage or BaseStorage()
        self.tg_suggests_cols = tg_suggests_cols

    def respond(self, message, source=None):
        # todo: support different triggers - not only messages, but calendar events as well
        if source is None:
            source = self.default_source
        user_id, message_text, metadata = self.standardize_input(source, message)
        user_object = self.get_user_object(user_id)
        old_user_object = copy.deepcopy(user_object)
        updated_user_object, response_text, suggests, response_commands = self.dialog_manager.respond(
            user_object, message_text, metadata
        )
        # todo: execute response_commands
        if updated_user_object != old_user_object:
            self.set_user_object(user_id, updated_user_object)
        response = self.standardize_output(source, message, response_text, response_commands, suggests)
        return response

    def get_user_object(self, user_id):
        if self.storage is None:
            return {}
        return self.storage.get(user_id)

    def set_user_object(self, user_id, user_object):
        if self.storage is None:
            raise NotImplementedError()
        self.storage.set(user_id, user_object)

    def standardize_input(self, source, message):
        metadata = {}
        if source == SOURCES.TELEGRAM:
            user_id = source + '__' + str(message.from_user.id)
            message_text = message.text
        elif source == SOURCES.ALICE:
            user_id = source + '__' + message['session']['user_id']
            message_text = message['request']['original_utterance']
            metadata['new_session'] = message.get('session', {}).get('new', False)
        elif source == SOURCES.TEXT:
            user_id = '0'
            message_text = message
        else:
            raise ValueError(SOURCES.unknown_source_error_message)
        return user_id, message_text, metadata

    def standardize_output(self, source, original_message, response_text, response_commands=None, suggests=None):
        has_exit_command = False
        if response_commands:
            for command in response_commands:
                if command == self.COMMAND_EXIT:
                    has_exit_command = True
                else:
                    raise NotImplementedError('Command "{}" is not implemented'.format(command))
        else:
            response_commands = []
        if source == SOURCES.TELEGRAM:
            response = {
                'text': response_text
            }
            if suggests:
                # todo: do smarter row width calculation
                row_width = min(self.tg_suggests_cols, len(suggests))
                response['reply_markup'] = telebot.types.ReplyKeyboardMarkup(row_width=row_width)
                response['reply_markup'].add(*[telebot.types.KeyboardButton(t) for t in suggests])
            else:
                response['reply_markup'] = telebot.types.ReplyKeyboardRemove(selective=False)
            return response
        elif source == SOURCES.ALICE:
            response = {
                "version": original_message['version'],
                "session": original_message['session'],
                "response": {
                    "end_session": has_exit_command,
                    "text": response_text
                }
            }
            if suggests:
                response['response']['buttons'] = [{'title': suggest, 'hide': True} for suggest in suggests]
            return response
        elif source == SOURCES.TEXT:
            response = response_text
            if suggests is not None and len(suggests) > 0:
                response = response + '\n' + ', '.join(['[{}]'.format(s) for s in suggests])
            if response_commands is not None and len(response_commands) > 0:
                response = response + '\n' + ', '.join(['{{{}}}'.format(c) for c in response_commands])
            return response, has_exit_command
        else:
            raise ValueError(SOURCES.unknown_source_error_message)



import copy
import json
import os

from . import database_utils


class BaseStorage:
    def __init__(self):
        self.dict = {}

    def get(self, key):
        return copy.deepcopy(self.dict.get(key, {}))

    def set(self, key, value):
        self.dict[key] = copy.deepcopy(value)


class FileBasedStorage(BaseStorage):
    def __init__(self, path='session_storage', multifile=True):
        super(FileBasedStorage, self).__init__()
        self.path = path
        self.multifile = multifile
        if not os.path.exists(path):
            if self.multifile:
                os.mkdir(path)
            else:
                self.dump_dict(path, {})

    def dump_dict(self, filename, data):
        with open(filename, 'w') as f:
            json.dump(data, f, indent=2)

    def load_dict(self, filename):
        with open(filename, 'r') as f:
            result = json.load(f)
        return result

    def get(self, key):
        if self.multifile:
            return self.load_dict(os.path.join(self.path, key))
        else:
            return self.load_dict(self.path).get(key, {})

    def set(self, key, value):
        if self.multifile:
            self.dump_dict(os.path.join(self.path, key), value)
        else:
            # todo: enable some concurrency guarantees
            data = self.load_dict(self.path)
            data[key] = value
            self.dump_dict(self.path, data)


class MongoBasedStorage(BaseStorage):
    KEY_NAME = 'key'
    VALUE_NAME = 'value'

    def __init__(self, database, collection_name='sessions'):
        super(MongoBasedStorage, self).__init__()
        self._collection = database.get_collection(collection_name)
        database_utils.ensure_mongo_index(index_name=self.KEY_NAME, collection=self._collection)

    def get(self, key):
        result = self._collection.find_one({self.KEY_NAME: key})
        if result is None:
            return {}
        return result.get(self.VALUE_NAME, {})

    def set(self, key, value):
        self._collection.update_one(
            {self.KEY_NAME: key},
            {'$set': {self.VALUE_NAME: value}},
            upsert=True
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tgalice",
    version="0.0.7",
    author="David Dale",
    author_email="dale.david@mail.ru",
    description="Yet another common wrapper for Telegram bots and Alice skills",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avidale/tgalice",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['pyTelegramBotAPI', 'textdistance', 'pyyaml', 'flask']
)



# -*- coding: utf-8 -*-
from __future__ import print_function

from . import dialog_manager, dialog_connector, session_storage, flask_server, nlu
from .dialog_manager.base import COMMANDS
from .message_logging import LoggedMessage
from tgalice.nlu import basic_nlu
